unit uVerify;
{$INCLUDE 'jedi.inc'}
{$INCLUDE 'vfy.inc'}

interface

uses
  SysUtils, Classes, Windows,
  uConsoleApp, uCompatibility, uSystem, uFiles,
  uHashList, uHashItem, uPCRE;

type
  TVerifyCommand = (vcVerifyDefault, vcVerify, {$IFNDEF USEDATETIME} vcVerifyFast, {$ENDIF} vcBuild, vcUpdate, {$IFNDEF USEDATETIME} vcUpdateFast, {$ENDIF} vcMatch);

  TProcessFileFunction = function (const BasePath, Path: string; Tag: integer): integer of object;
  TAfterProcessFunction = procedure (var ProcessResult: integer; Tag: integer) of object;

  TIncludeExcludeMask = (iemInclude, iemExclude, iemIncludeRegExp, iemExcludeRegExp);

  EVerifyAppError = class(Exception);

  TVerifyApp = class(TConsoleApp)
    protected
      function AppTitle: string; override;
      function AppVersion: string; override;
      function DoExecute: integer; override;
      procedure DoConsoleCtrl(const CtrlType: DWORD; var Handled: boolean); override;
      procedure ShowSyntax(const Error: string = ''); override;
      procedure ParamInitialize; override;
    private
      fVerifyCommand: TVerifyCommand;
      fHashesFileName: string;
      fOutputFileName: string;
      fErrorLogFileName: string;
      fMatchListFileName: string;
      fBaseDirectories: TStringList;
      fRecursive: boolean;
      fJunctions: boolean;
      fHashList: THashList;
      fLastProgressTicks: DWORD;
      fShowProgress: boolean;
      fIncludeExcludeList: TStringList;
      fControlEvent: DWORD;
      {$IFDEF CASESENSITIVESUPPORT}
      fCaseSensitive: boolean;
      {$ENDIF}
      {$IFDEF OVERRIDEWOW64}
      fWow64RedirectionOldValueDONOTCHANGE: Pointer;
      fDisableWow64: boolean;
      {$ENDIF}
      {$IFDEF USEDATETIME}
      fUseTimestamp: boolean;
      fTimestampPrecision: integer;
      {$ENDIF}
      fHideStates: THashStates;
    protected
      function ParseCommandLine: integer; virtual;
      procedure FixUpBaseDirectories; virtual;
      procedure HashingProgress(Sender: TObject; Done, Total: int64); virtual;
      function ErrorInFile(E: Exception; const BasePath, Path: string): integer; virtual;
      function VerifyOutput(Item: THashItem): string; virtual;
      function ProcessFiles(const BasePath, Path: string; ProcessFileFunction: TProcessFileFunction; Tag: integer): integer; virtual;
      function ProcessFile_Build(const BasePath, Path: string; Tag: integer): integer; virtual;
      function ProcessFile_Verify(const BasePath, Path: string; Tag: integer): integer; virtual;
      {$IFNDEF USEDATETIME}
      function ProcessFile_VerifyFast(const BasePath, Path: string; Tag: integer): integer; virtual;
      {$ENDIF}
      function ProcessFile_Update(const BasePath, Path: string; Tag: integer): integer; virtual;
      {$IFNDEF USEDATETIME}
      function ProcessFile_UpdateFast(const BasePath, Path: string; Tag: integer): integer; virtual;
      {$ENDIF}
      procedure AfterProcess_Verify(var ProcessResult: integer; Tag: integer); virtual;
      procedure AfterProcess_Update(var ProcessResult: integer; Tag: integer); virtual;
      function InternalVerify(ProcessFileFunction: TProcessFileFunction; AfterProcessFunction: TAfterProcessFunction; Tag: integer): integer; virtual;
      procedure ApplyIncludeExcludeToHashList(HashList: THashList); virtual;
      function IsFileWanted(const Path: string): boolean; virtual;
      procedure CheckControlEvent; virtual;
      function HandleCtrlBreakError(const Msg: string): boolean; virtual;
      procedure LoadHashFile(const FileName: string; HashList: THashList); virtual;
      {$IFDEF CASESENSITIVESUPPORT}
      function FindFirst(const Path: string; Attr: Integer; var SR: TSearchRec): Integer; virtual;
      {$ENDIF}
      {$IFDEF OVERRIDEWOW64}
        procedure ChangeWow64Redirection(Active: boolean); virtual;
        {$IFNDEF OVERRIDEWOW64GLOBALLY}
          {$IFDEF DELPHI6_UP}
            {$MESSAGE ERROR 'OVERRIDEWOW64 is not done yet.'}
          {$ELSE}
            'OVERRIDEWOW64 is not done yet.'
          {$ENDIF}
        {$ENDIF}
      {$ENDIF}
      property HashList: THashList read fHashList;
      property IncludeExcludeList: TStringList read fIncludeExcludeList;
      property ControlEvent: DWORD read fControlEvent write fControlEvent;
    public
      constructor Create; override;
      destructor Destroy; override;
      function Build: integer; virtual;
      function Verify: integer; virtual;
      {$IFNDEF USEDATETIME}
      function VerifyFast: integer; virtual;
      {$ENDIF}
      function Update: integer; virtual;
      {$IFNDEF USEDATETIME}
      function UpdateFast: integer; virtual;
      {$ENDIF}
      function Match: integer; virtual;
      procedure ClearIncludeExcludeMasks; virtual;
      procedure AddIncludeExcludeMask(const Mask: string; MaskType: TIncludeExcludeMask); virtual;
      property VerifyCommand: TVerifyCommand read fVerifyCommand write fVerifyCommand;
      property HashesFileName: string read fHashesFileName write fHashesFileName;
      property OutputFileName: string read fOutputFileName write fOutputFileName;
      property ErrorLogFileName: string read fErrorLogFileName write fErrorLogFileName;
      property MatchListFileName: string read fMatchListFileName write fMatchListFileName;
      property BaseDirectories: TStringList read fBaseDirectories;
      property Recursive: boolean read fRecursive write fRecursive;
      property ShowProgress: boolean read fShowProgress write fShowProgress;
      property Junctions: boolean read fJunctions write fJunctions;
      {$IFDEF CASESENSITIVESUPPORT}
      property CaseSensitive: boolean read fCaseSensitive write fCaseSensitive;
      {$ENDIF}
      {$IFDEF OVERRIDEWOW64}
      property DisableWow64: boolean read fDisableWow64 write fDisableWow64;
      {$ENDIF}
      {$IFDEF USEDATETIME}
      property UseTimestamp: boolean read fUseTimestamp write fUseTimestamp;
      property TimestampPrecision: integer read fTimestampPrecision write fTimestampPrecision; // Milliseconds
      {$ENDIF}
      property HideStates: THashStates read fHideStates write fHideStates;
    end;

const
  EXITCODE_OK = 0;
  EXITCODE_CHANGED_FILES = 1;
  EXITCODE_UNKNOWN_COMMAND = 101;
  EXITCODE_NO_DATA = 102;
  EXITCODE_ERROR_IN_FILES = 103;
  EXITCODE_BREAK = 253;
  EXITCODE_SYNTAX = 254;

implementation

{$IFDEF OVERRIDEWOW64GLOBALLY}
  {$IFDEF DELPHI6_UP}
    {$MESSAGE WARN 'Using OVERRIDEWOW64GLOBALLY. Don''t use LoadLibrary during Build/Verify/Update!'}
  {$ENDIF}
{$ENDIF}

const
  PROGRESS_EVENT_TICKS = 1000;

const
  INCEXCTYPE_INCLUDE = 'I';
  INCEXCTYPE_EXCLUDE = 'X';

const
  TAG_BUILD = 1;
  TAG_VERIFY = 2;
  TAG_VERIFYFAST = 3;
  TAG_UPDATE = 4;
  TAG_UPDATEFAST = 5;

const
  CTRL_NO_EVENT = $ffffffff;

const
  PROGRESS_CLEAR = '       ';

{ TVerifyApp }

constructor TVerifyApp.Create;
begin
  fBaseDirectories := TStringList.Create;
  fHashList := THashList.Create;
  fIncludeExcludeList := TStringList.Create;
  fControlEvent := CTRL_NO_EVENT;
  {$IFDEF OVERRIDEWOW64}
  fWow64RedirectionOldValueDONOTCHANGE := nil;
  {$ENDIF}
  inherited;
end;

destructor TVerifyApp.Destroy;
begin
  FreeAndNil(fBaseDirectories);
  FreeAndNil(fHashList);
  ClearIncludeExcludeMasks;
  FreeAndNil(fIncludeExcludeList);
  inherited;
end;

function TVerifyApp.AppTitle: string;
begin
  Result := 'Verify';
end;

function TVerifyApp.AppVersion: string;
begin
  Result := {$INCLUDE 'vfy.version'} ;
end;

procedure TVerifyApp.ShowSyntax(const Error: string);
begin
  inherited;
  WriteColored(ccWhite, '[<arg>] [<arg>] ... <hashfile> [<dir>] [<dir>] ...'); Writeln;
  Writeln;
  WriteColored(ccYellow, 'Commands:'); Writeln;
  WriteColored(ccWhite, ' -h, -?'); Writeln(' ......... Show this help screen.');
  Writeln;
  WriteColored(ccWhite, ' -b'); Writeln(' ............. Build a new hash list.');
  WriteColored(ccWhite, ' -v'); Writeln(' ............. Verify an existing hash list (default).');
  {$IFNDEF USEDATETIME}
  WriteColored(ccWhite, ' -vf'); Writeln(' ............ Fast verify (skip hashing new files).');
  {$ENDIF}
  WriteColored(ccWhite, ' -u'); Writeln(' ............. Verify and update an existing hash list.');
  {$IFNDEF USEDATETIME}
  WriteColored(ccWhite, ' -uf'); Writeln(' ............ Fast update (skip verify of existing files).');
  {$ENDIF}
  WriteColored(ccWhite, ' -m <file>'); Writeln(' ...... Match two hash lists, <file> being the older.');
  Writeln;
  WriteColored(ccYellow, 'Matching specifiers:'); Writeln;
  WriteColored(ccWhite, ' -j[-]'); Writeln(' .......... [Don''t] Scan junctions and reparse points (default: -j).');
  WriteColored(ccWhite, ' -r[-]'); Writeln(' .......... [Don''t] Recursively scan subdirectories (default: -r).');
  {$IFDEF USEDATETIME}
  WriteColored(ccWhite, ' -t[-]'); Writeln(' .......... [Don''t] Use file modification timestamp (default: -t).');
  WriteColored(ccWhite, ' -tp <msec>'); Writeln(' ..... Use <msec> millisecond precision for comparison (default: 0).');
  {$ENDIF}
  {$IFDEF CASESENSITIVESUPPORT}
  WriteColored(ccWhite, ' -cs[-]'); Writeln(' ......... [Don''t] Use case-sensitive file functions (default:: -cs-).');
  {$ENDIF}
  {$IFDEF OVERRIDEWOW64}
  WriteColored(ccWhite, ' -wow64[-]'); Writeln(' ...... [Disable] Wow64 filesystem redirection (default: -wow64).');
  {$ENDIF}
  Writeln;
  WriteColored(ccYellow, 'Source files:'); Writeln;
  WriteColored(ccWhite, ' -d <dir>'); Writeln(' ....... Add directory <dir> to scanned directories.');
  WriteColored(ccWhite, ' -i <wildcard>'); Writeln('... Include files whose name matches <wildcard>.');
  WriteColored(ccWhite, ' -x <wildcard>'); Writeln('... Exclude files whose name matches <wildcard>.');
  WriteColored(ccWhite, ' -ir <regexp>'); Writeln('.... Include files whose full name matches <regexp>.');
  WriteColored(ccWhite, ' -xr <regexp>'); Writeln('.... Exclude files whose full name matches <regexp>.');
  Writeln;
  WriteColored(ccYellow, 'Output settings:'); Writeln;
  WriteColored(ccWhite, ' -p[-]'); Writeln(' .......... [Don''t] Show progress.');
  WriteColored(ccWhite, ' -e <file>'); Writeln(' ...... Write error log to file.');
  WriteColored(ccWhite, ' -o <file>'); Writeln(' ...... Write verify/build/match results to <file>.');
  WriteColored(ccWhite, ' -sm[-]'); Writeln(' ......... [Don''t] Show matched (non-changed) files (default: -sm-).');
  WriteColored(ccWhite, ' -sa[-]'); Writeln(' ......... [Don''t] Show added (new) files (default: -sa).');
  WriteColored(ccWhite, ' -sc[-]'); Writeln(' ......... [Don''t] Show changed files (default: -sc).');
  WriteColored(ccWhite, ' -sd[-]'); Writeln(' ......... [Don''t] Show deleted files (default: -sd).');
  WriteColored(ccWhite, ' -sp[-]'); Writeln(' ......... [Don''t] Show copied files (default: -sp-).');
  WriteColored(ccWhite, ' -sr[-]'); Writeln(' ......... [Don''t] Show renamed files (default: -sr-).');
  Writeln;
  {$IFDEF DELPHICZBANNER}
  WriteColored(ccYellow, 'This utility was entered in a Delphi.cz competition. You can read the details'); Writeln;
  WriteColored(ccYellow, 'of this competition here:'); Writeln;
  WriteColored(ccLightCyan, 'http://delphi.cz/post/Soutez-o-RAD-Studio-2007-Architekt.aspx'); Writeln;
  Writeln;
  {$ENDIF}
  //Writeln;
end;

procedure TVerifyApp.ParamInitialize;
begin
  inherited;
  VerifyCommand := vcVerifyDefault;
  HashesFileName := '';
  OutputFileName := '';
  ErrorLogFileName := '';
  MatchListFileName := '';
  BaseDirectories.Clear;
  Recursive := True;
  Junctions := True;
  ShowProgress := not StdOutRedirected;
  {$IFDEF CASESENSITIVESUPPORT}
  CaseSensitive := False;
  {$ENDIF}
  {$IFDEF OVERRIDEWOW64}
  DisableWow64 := False;
  {$ENDIF}
  {$IFDEF USEDATETIME}
  UseTimestamp := True;
  TimestampPrecision := 0;
  {$ENDIF}
  HideStates := [hsOK, hsCopiedFrom, hsMovedTo];
end;

function TVerifyApp.DoExecute: integer;
begin
  // Read command line arguments
  ParamInitialize;
  Result := ParseCommandLine;
  if Result <> EXITCODE_OK then
    Exit;
  try
    case VerifyCommand of
      vcBuild:
        Result := Build;
      vcVerify, vcVerifyDefault:
        Result := Verify;
      {$IFNDEF USEDATETIME}
      vcVerifyFast:
        Result := VerifyFast;
      {$ENDIF}
      vcUpdate:
        Result := Update;
      {$IFNDEF USEDATETIME}
      vcUpdateFast:
        Result := UpdateFast;
      {$ENDIF}
      vcMatch:
        Result := Match;
      else
        Result := EXITCODE_UNKNOWN_COMMAND;
      end;
  except
    on E: EControlC do
      Result := EXITCODE_BREAK;
    end;
end;

function TVerifyApp.ParseCommandLine: integer;
const ONLY_ONE_COMMAND = 'Only one of -b/-v/-vf/-u/-uf/-m commands may be present.';

  procedure SyntaxError(const Msg: string); overload;
    begin
      ShowSyntax(Msg);
      Abort;
    end;
  procedure SyntaxError(const Msg: string; const Params: array of const); overload;
    begin
      SyntaxError(Format(Msg, Params));
    end;

var Param: string;
    iParam: integer;
begin
  try
    while ParamGetNext(Param) do
      begin
      if Param[1] = '-' then
        begin
        if (Param = '-h') or (Param = '-?') then
          SyntaxError('')
        else if Param = '-v' then
          if VerifyCommand = vcVerifyDefault then
            VerifyCommand := vcVerify
          else
            SyntaxError(ONLY_ONE_COMMAND)
        {$IFNDEF USEDATETIME}
        else if Param = '-vf' then
          if VerifyCommand = vcVerifyDefault then
            VerifyCommand := vcVerifyFast
          else
            SyntaxError(ONLY_ONE_COMMAND)
        {$ENDIF}
        else if Param = '-b' then
          if VerifyCommand = vcVerifyDefault then
            VerifyCommand := vcBuild
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-u' then
          if VerifyCommand = vcVerifyDefault then
            VerifyCommand := vcUpdate
          else
            SyntaxError(ONLY_ONE_COMMAND)
        {$IFNDEF USEDATETIME}
        else if Param = '-uf' then
          if VerifyCommand = vcVerifyDefault then
            VerifyCommand := vcUpdateFast
          else
            SyntaxError(ONLY_ONE_COMMAND)
        {$ENDIF}
        else if Param = '-m' then
          if VerifyCommand = vcVerifyDefault then
            if ParamGetNext(Param) then
              if FileExists(Param) then
                begin
                VerifyCommand := vcMatch;
                MatchListFileName := Param;
                end
              else
                SyntaxError('Matchfile not found for -m command.')
            else
              SyntaxError('File name must be provided for -m command.')
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-r' then
          Recursive := True
        else if Param = '-r-' then
          Recursive := False
        else if Param = '-p' then
          ShowProgress := True
        else if Param = '-p-' then
          ShowProgress := False
        else if Param = '-j' then
          Junctions := True
        else if Param = '-j-' then
          Junctions := False
        {$IFDEF USEDATETIME}
        else if Param = '-t' then
          UseTimestamp := True
        else if Param = '-t-' then
          UseTimestamp := False
        else if Param = '-tp' then
          if ParamGetNext(Param) then
            begin
            iParam := StrToIntDef(Param, -1);
            if iParam >= 0 then
              TimestampPrecision := iParam
            else
              SyntaxError('Precision must be an integer number, not "%s".', [Param]);
            end
          else
            SyntaxError('Precision must be provided with -tp argument.')
        {$ENDIF}
        {$IFDEF CASESENSITIVESUPPORT}
        else if Param = '-cs' then
          CaseSensitive := True
        else if Param = '-cs-' then
          CaseSensitive := False
        {$ENDIF}
        {$IFDEF OVERRIDEWOW64}
        else if Param = '-wow64' then
          DisableWow64 := False
        else if Param = '-wow64-' then
          DisableWow64 := True
        {$ENDIF}
        else if Param = '-sm' then
          HideStates := HideStates - [hsOK]
        else if Param = '-sm-' then
          HideStates := HideStates + [hsOK]
        else if Param = '-sa' then
          HideStates := HideStates - [hsNew]
        else if Param = '-sa-' then
          HideStates := HideStates + [hsNew]
        else if Param = '-sc' then
          HideStates := HideStates - [hsChanged]
        else if Param = '-sc-' then
          HideStates := HideStates + [hsChanged]
        else if Param = '-sd' then
          HideStates := HideStates - [hsDeleted]
        else if Param = '-sd-' then
          HideStates := HideStates + [hsDeleted]
        else if Param = '-sp' then
          HideStates := HideStates - [hsCopiedFrom]
        else if Param = '-sp-' then
          HideStates := HideStates + [hsCopiedFrom]
        else if Param = '-sr' then
          HideStates := HideStates - [hsMovedTo]
        else if Param = '-sr-' then
          HideStates := HideStates + [hsMovedTo]
        else if Param = '-i' then
          if ParamGetNext(Param) then
            AddIncludeExcludeMask(Param, iemInclude)
          else
            SyntaxError('Mask must be provided with -i argument.')
        else if Param = '-x' then
          if ParamGetNext(Param) then
            AddIncludeExcludeMask(Param, iemExclude)
          else
            SyntaxError('Mask must be provided with -x argument.')
        else if Param = '-ir' then
          if ParamGetNext(Param) then
            AddIncludeExcludeMask(Param, iemIncludeRegExp)
          else
            SyntaxError('Regular expression must be provided with -ir argument.')
        else if Param = '-xr' then
          if ParamGetNext(Param) then
            AddIncludeExcludeMask(Param, iemExcludeRegExp)
          else
            SyntaxError('Regular expression ask must be provided with -xr argument.')
        else if Param = '-d' then
          if ParamGetNext(Param) then
            BaseDirectories.Add(Param)
          else
            SyntaxError('Directory name must be provided with -d argument.')
        else if Param = '-o' then
          if ParamGetNext(Param) then
            if Param = '-' then
              OutputFileName := ''
            else
              OutputFileName := Param
          else
            SyntaxError('File name must be provided with -o argument.')
        else if Param = '-e' then
          if ParamGetNext(Param) then
            ErrorLogFileName := Param
          else
            SyntaxError('File name must be provided with -e argument.')
        else
          SyntaxError('Unknown argument "%s".', [Param])
        end
      else
        if HashesFileName = '' then
          HashesFileName := Param
        else
          BaseDirectories.Add(Param);
      end;
    if HashesFileName = '' then
      SyntaxError('Hashfile must be provided.');
    if BaseDirectories.Count = 0 then
      BaseDirectories.Add('.');
    Result := EXITCODE_OK;
  except
    on EAbort do
      Result := EXITCODE_SYNTAX;
    end;
end;

procedure TVerifyApp.FixUpBaseDirectories;
var i: integer;
    Dir: string;
begin
  for i := 0 to Pred(BaseDirectories.Count) do
    begin
    Dir := BaseDirectories[i];
    if Dir = '' then
      Dir := IncludeTrailingPathDelimiter(GetCurrentDir)
    else
      Dir := IncludeTrailingPathDelimiter(ExpandFileName(Dir));
    {$IFDEF LONGPATHS}
      if Dir <> '' then
        if Copy(Dir, 1, 2) = '\\' then
          Dir := '\\?\UNC\' + Copy(Dir, 3, MaxInt)
        else
          Dir := '\\?\' + Dir;
    {$ENDIF}
    BaseDirectories[i] := Dir;
    end;
end;

function TVerifyApp.ErrorInFile(E: Exception; const BasePath, Path: string): integer;
var Error: string;
begin
  Result := EXITCODE_ERROR_IN_FILES;
  Error := Format(#13#10'ERROR %s for file "%s%s": %s', [E.ClassName, BasePath, Path, E.Message]);
  if ErrorLogFileName <> '' then
    Log(ErrorLogFileName, Error);
  ShowError(Error);
end;

function TVerifyApp.VerifyOutput(Item: THashItem): string;
begin
  Result := '';
  case Item.State of
    hsOK:
      Result := 'MATCHED'#9 + Item.AsString;
    hsNew:
      Result := 'ADDED'#9 + Item.AsString;
    hsChanged:
      Result := 'CHANGED'#9 + Item.AsString;
    hsDeleted:
      Result := 'DELETED'#9 + Item.AsString;
    hsCopiedFrom:
      if Item.Related = nil then
        Result := 'COPIED'#9 + Item.AsString // This should never happen
      else
        Result := 'COPIED'#9'"' + Item.Path + '"'#9 + ' from ' + #9'"' + Item.Related.Path + '"';
    hsMovedTo:
      if Item.Related = nil then
        Result := 'RENAMED'#9 + Item.AsString // This should never happen
      else
        Result := 'RENAMED'#9'"' + Item.Path + '"'#9 + ' to ' + #9'"' + Item.Related.Path + '"';
    end;
end;

procedure TVerifyApp.LoadHashFile(const FileName: string; HashList: THashList);
begin
  HashList.Clear;
  if ShowProgress then
    begin
    Write('Loading hash file '); WriteColored(ccWhite, FileName); Write(' ... ');
    end;
  HashList.LoadFromFile(FileName, '', HashingProgress);
  ApplyIncludeExcludeToHashList(HashList);
  if ShowProgress then
    Writeln(PROGRESS_CLEAR);
end;

function TVerifyApp.Build: integer;
var i, Code: integer;
begin
  Result := EXITCODE_NO_DATA;
  ControlEvent := CTRL_NO_EVENT;
  {$IFDEF OVERRIDEWOW64GLOBALLY}
  if DisableWow64 then
    ChangeWow64Redirection(False);
  try
  {$ENDIF}
  if FileExists(HashesFileName) then
    DeleteFile(PChar(HashesFileName));
  FixUpBaseDirectories;
  HashList.Clear;
  try
    for i := 0 to Pred(BaseDirectories.Count) do
      begin
      Code := ProcessFiles(BaseDirectories[i], '', ProcessFile_Build, TAG_BUILD);
      if Result = EXITCODE_NO_DATA then
        Result := Code
      else if Code <> EXITCODE_OK then
        Result := Code;
      end;
  except
    on E: EControlC do
      if HandleCtrlBreakError(E.Message) then
        Result := EXITCODE_BREAK
      else
        Raise;
    end;
  //if Result = EXITCODE_OK then
    begin
    if ShowProgress then
      Writeln ('Sorting results...');
    HashList.SortByPath;
    HashList.SaveToFile(HashesFileName);
    end;
  {$IFDEF OVERRIDEWOW64GLOBALLY}
  finally
    if DisableWow64 then
      ChangeWow64Redirection(True);
    end;
  {$ENDIF}
end;

function TVerifyApp.ProcessFile_Build(const BasePath, Path: string; Tag: integer): integer;
var HashItem: THashItem;
begin
  Result := EXITCODE_OK;
  try
    HashItem := THashItem.Create;
    try
      HashItem.ReadFromFile(BasePath, Path, HashingProgress {$IFDEF CASESENSITIVESUPPORT} , CaseSensitive {$ENDIF} );
      HashItem.State := hsNew;
      HashList.Add(HashItem);
      {$IFDEF SHOWRESULTSIMMEDIATELY}
      if HashesFileName <> '' then
        Log(HashesFileName, HashItem.AsString);
      {$ENDIF}
    except
      HashItem.Free;
      Raise;
      end;
  except
    on E: EControlC do
      Raise;
    on E: Exception do
      Result := ErrorInFile(E, BasePath, Path);
    end;
end;

function TVerifyApp.InternalVerify(ProcessFileFunction: TProcessFileFunction; AfterProcessFunction: TAfterProcessFunction; Tag: integer): integer;
var i, Code: integer;
    Msg: string;
begin
  Result := EXITCODE_NO_DATA;
  ControlEvent := CTRL_NO_EVENT;
  {$IFDEF OVERRIDEWOW64GLOBALLY}
  if DisableWow64 then
    ChangeWow64Redirection(False);
  try
  {$ENDIF}
  if OutputFileName <> '' then
    if FileExists(OutputFileName) then
      DeleteFile(PChar(OutputFileName));
  FixUpBaseDirectories;
  LoadHashFile(HashesFileName, HashList);
  if ShowProgress then
    Writeln('Sorting hash list...');
  HashList.SortByPath;
  for i := 0 to Pred(HashList.Count) do
    HashList[i].State := hsDeleted;
  try
    for i := 0 to Pred(BaseDirectories.Count) do
      begin
      Code := ProcessFiles(BaseDirectories[i], '', ProcessFileFunction, Tag);
      if Result = EXITCODE_NO_DATA then
        Result := Code
      else if Code <> EXITCODE_OK then
        Result := Code;
      end;
  except
    on E: EControlC do
      if HandleCtrlBreakError(E.Message) then
        Result := EXITCODE_BREAK
      else
        Raise;
    end;
  if Assigned(AfterProcessFunction) then
    AfterProcessFunction(Result, Tag);
  if OutputFileName <> '' then
    if FileExists(OutputFileName) then
      DeleteFile(PChar(OutputFileName));
  HashList.SortByPath;
  for i := 0 to Pred(HashList.Count) do
    if not (HashList[i].State in HideStates) then
      begin
      if (Result = EXITCODE_OK) or (Result = EXITCODE_NO_DATA) then
        Result := EXITCODE_CHANGED_FILES;
      Msg := VerifyOutput(HashList[i]);
      //Writeln(Msg);
      //if OutputFileName <> '' then
        Log(OutputFileName, Msg);
      end;
  {$IFDEF OVERRIDEWOW64GLOBALLY}
  finally
    if DisableWow64 then
      ChangeWow64Redirection(True);
    end;
  {$ENDIF}
end;

function TVerifyApp.Verify: integer;
begin
  Result := InternalVerify(ProcessFile_Verify, AfterProcess_Verify, TAG_VERIFY);
end;

function TVerifyApp.ProcessFile_Verify(const BasePath, Path: string; Tag: integer): integer;
var HashItem: THashItem;
    Index: integer;
begin
  Result := EXITCODE_OK;
  try
    HashItem := THashItem.Create;
    try
      HashItem.ReadFromFile(BasePath, Path, HashingProgress {$IFDEF CASESENSITIVESUPPORT} , CaseSensitive {$ENDIF} );
      if HashList.FindByPath(Path, Index) then
        if (HashList[Index].Size = HashItem.Size) and (HashList[Index].Hash = HashItem.Hash) then
          begin
          HashList[Index].State := hsOK;
          FreeAndNil(HashItem);
          end
        else
          begin
          HashItem.State := hsChanged;
          //HashList[Index].Free; // Free will be done by THashList itself due to OwnObjects=True
          HashList[Index] := HashItem;
          {$IFDEF SHOWRESULTSIMMEDIATELY}
          if OutputFileName <> '' then
            Log(OutputFileName, VerifyOutput(HashItem));
          {$ENDIF}
          HashItem := nil;
          Result := EXITCODE_CHANGED_FILES;
          end
      else
        begin
        HashItem.State := hsNew;
        HashList.Insert(Index, HashItem);
        {$IFDEF SHOWRESULTSIMMEDIATELY}
        if OutputFileName <> '' then
          Log(OutputFileName, VerifyOutput(HashItem));
        {$ENDIF}
        HashItem := nil;
        Result := EXITCODE_CHANGED_FILES;
        end;
    except
      HashItem.Free;
      Raise;
      end;
  except
    on E: EControlC do
      Raise;
    on E: Exception do
      Result := ErrorInFile(E, BasePath, Path);
    end;
end;

procedure TVerifyApp.AfterProcess_Verify(var ProcessResult: integer; Tag: integer);
var i, Index: integer;
begin
  //if ProcessResult in [EXITCODE_OK, EXITCODE_CHANGED_FILES] then
    begin
    HashList.SortByHash;
    for i := 0 to Pred(HashList.Count) do
      if HashList[i].State in [hsNew, hsDeleted] then
        if HashList.FindByHash(HashList[i].Hash, HashList[i].Size, Index) then
          while (Index < HashList.Count) and (HashItemCompareByHash(HashList[i], HashList[Index]) = 0) do
            if i = Index then
              Inc(Index)
            else
              begin
              HashList[i].Related := HashList[Index];
              case HashList[i].State of
                hsNew:
                  HashList[i].State := hsCopiedFrom;
                hsDeleted:
                  HashList[i].State := hsMovedTo;
                end;
              Break;
              end;
    end;
end;

{$IFNDEF USEDATETIME}
function TVerifyApp.VerifyFast: integer;
begin
  Result := InternalVerify(ProcessFile_VerifyFast, nil, TAG_VERIFYFAST);
end;

function TVerifyApp.ProcessFile_VerifyFast(const BasePath, Path: string; Tag: integer): integer;
var Index: integer;
    HashItem: THashItem;
    Error: string;
begin
  try
    if not HashList.FindByPath(Path, Index) then
      begin
      HashItem := THashItem.Create;
      HashItem.BasePath := BasePath;
      HashItem.Path := Path;
      HashItem.State := hsNew;
      HashList.Insert(Index, HashItem);
      {$IFDEF SHOWRESULTSIMMEDIATELY}
      if OutputFileName <> '' then
        Log(OutputFileName, VerifyOutput(HashItem));
      {$ENDIF}
      Result := EXITCODE_OK;
      end
    else
      Result := ProcessFile_Verify(BasePath, Path, Tag);
  except
    on E: EControlC do
      Raise;
    on E: Exception do
      Result := ErrorInFile(E, BasePath, Path);
    end;
end;
{$ENDIF}

function TVerifyApp.Update: integer;
begin
  Result := InternalVerify(ProcessFile_Update, AfterProcess_Update, TAG_UPDATE);
end;

function TVerifyApp.ProcessFile_Update(const BasePath, Path: string; Tag: integer): integer;
{$IFDEF USEDATETIME}
const MILLISECONDS_PER_DAY = 24 * 60 * 60 * 1000;
var Index: integer;
{$ENDIF}
begin
  {$IFDEF USEDATETIME}
  try
    if UseTimestamp
       and HashList.FindByPath(Path, Index)
       and (Abs(HashList[Index].Date - FileGetDateTime(BasePath + Path)) <= (TimestampPrecision * 1/MILLISECONDS_PER_DAY))
    then
      begin
      HashList[Index].State := hsOK;
      Result := EXITCODE_OK;
      end
    else
  {$ENDIF}
  Result := ProcessFile_Verify(BasePath, Path, Tag);
  if Result = EXITCODE_CHANGED_FILES then
    begin
    //HashList.DeleteByState(hsDeleted);
    //HashList.DeleteByState(hsRenamedTo);
    HashList.SaveToFile(HashesFileName);
    end;
  {$IFDEF USEDATETIME}
  except
    on E: EControlC do
      Raise;
    on E: Exception do
      Result := ErrorInFile(E, BasePath, Path);
    end;
  {$ENDIF}
end;

procedure TVerifyApp.AfterProcess_Update(var ProcessResult: integer; Tag: integer);
begin
  //if ProcessResult in [EXITCODE_OK, EXITCODE_CHANGED_FILES] then
    begin
    HashList.SortByPath;
    HashList.DeleteByState(hsDeleted);
    HashList.DeleteByState(hsMovedTo);
    HashList.SaveToFile(HashesFileName);
    end;
end;

{$IFNDEF USEDATETIME}
function TVerifyApp.UpdateFast: integer;
begin
  Result := InternalVerify(ProcessFile_UpdateFast, nil, TAG_UPDATEFAST);
  if Result = EXITCODE_CHANGED_FILES then
    begin
    HashList.DeleteByState(hsDeleted);
    HashList.DeleteByState(hsRenamedTo);
    HashList.SaveToFile(HashesFileName);
    end;
end;

function TVerifyApp.ProcessFile_UpdateFast(const BasePath, Path: string; Tag: integer): integer;
var Index: integer;
    Error: string;
begin
  try
    if HashList.FindByPath(Path, Index) then
      begin
      HashList[Index].State := hsOK;
      Result := EXITCODE_OK;
      end
    else
      Result := ProcessFile_Update(BasePath, Path, Tag);
  except
    on E: EControlC do
      Raise;
    on E: Exception do
      Result := ErrorInFile(E, BasePath, Path);
    end;
end;
{$ENDIF}

function TVerifyApp.Match: integer;
  // HashesFileName = the new hash list
  // MatchListFileName = the old hash list
var MatchList: THashList;
    //Item: THashItem;
    Msg: string;
    i, Index: integer;
    Done, Total: int64;

  procedure Step;
    begin
      Inc(Done);
      HashingProgress(Self, Done, Total);
    end;

begin
  Result := EXITCODE_OK;
  ControlEvent := CTRL_NO_EVENT;
  if OutputFileName <> '' then
    if FileExists(OutputFileName) then
      DeleteFile(PChar(OutputFileName));
  LoadHashFile(HashesFileName, HashList);
  MatchList := THashList.Create;
  try
    // hsOK = stejna cesta i hash
    // hsCopied = ruzna cesta, stejny hash
    // hsChanged = stejna cesta, ruzny hash (ale jen pokud neni hsCopied)
    // hsNew = nenalezen ve starem hash listu
    // hsDeleted = nenalezen v novem hash listu
    LoadHashFile(MatchListFileName, MatchList);
    if ShowProgress then
      Write('Comparing hash lists...');
    Total := 1 * MatchList.Count + 2 * HashList.Count;
    Done := -1;
    Step;
    {
    for i := 0 to Pred(HashList.Count) do
      begin
      HashList[i].BasePath := '*New*';
      Step;
      end;
    }
    for i := 0 to Pred(MatchList.Count) do
      begin
      MatchList[i].State := hsDeleted;
      //MatchList[i].BasePath := '*Old*';
      Step;
      end;
    // 1. Match by path
    MatchList.SortByPath;
    for i := 0 to Pred(HashList.Count) do
      begin
      if MatchList.FindByPath(HashList[i].Path, Index) then
        begin
        HashList[i].State := hsChanged;
        HashList[i].Related := MatchList[Index];
        while (Index < MatchList.Count) and (HashItemCompareByPath(HashList[i], MatchList[Index]) = 0) do
          begin
          if HashItemCompareByHash(HashList[i], MatchList[Index]) = 0 then
            begin
            if HashList[i].State <> hsOK then
              begin
              HashList[i].State := hsOK;
              HashList[i].Related := MatchList[Index];
              end;
            MatchList[Index].State := hsOK;
            end
          else
            begin
            if MatchList[Index].State = hsDeleted then
              MatchList[Index].State := hsChanged;
            end;
          Inc(Index);
          end;
        end
      else
        begin
        HashList[i].Related := nil;
        HashList[i].State := hsNew;
        end;
      Step;
      end;
    // 2. Match not-matched by hash
    MatchList.SortByHash;
    for i := 0 to Pred(HashList.Count) do
      begin
      if (HashList[i].State <> hsOK) and (HashList[i].State <> hsChanged) then
        if MatchList.FindByHash(HashList[i].Hash, HashList[i].Size, Index) then
          begin
          HashList[i].Related := MatchList[Index];
          HashList[i].State := hsCopiedFrom;
          while (Index < MatchList.Count) and (HashItemCompareByHash(HashList[i], MatchList[Index]) = 0) do
            begin
            if MatchList[Index].State <> hsOK then
              begin
              MatchList[Index].State := hsMovedTo;
              MatchList[Index].Related := HashList[i];
              end;
            Inc(Index);
            end;
          end;
        Step;
        end;
    // 3. Print out the results, groupped by State
    if ShowProgress then
      begin
      Writeln(PROGRESS_CLEAR);
      Writeln('Sorting results...');
      end;
    HashList.SortByPath;
    MatchList.SortByPath;
    if ShowProgress then
      Writeln('Writing results...');
    Log(OutputFileName, '* New hashlist *');
    for i := 0 to Pred(HashList.Count) do
      if not (HashList[i].State in HideStates) then
        begin
        Msg := VerifyOutput(HashList[i]);
        if Msg <> '' then
          begin
          //Writeln(Msg);
          //if OutputFileName <> '' then
            Log(OutputFileName, Msg);
          Result := EXITCODE_CHANGED_FILES;
          end;
        end;
    Log(OutputFileName, '');
    Log(OutputFileName, '* Old hashlist *');
    for i := 0 to Pred(MatchList.Count) do
      if not (MatchList[i].State in (HideStates + [hsOK, hsNew, hsChanged])) then
        begin
        Msg := VerifyOutput(MatchList[i]);
        if Msg <> '' then
          begin
          //Writeln(Msg);
          //if OutputFileName <> '' then
            Log(OutputFileName, Msg);
          Result := EXITCODE_CHANGED_FILES;
          end;
        end;
  finally
    MatchList.Free;
    end;
end;

{$IFDEF OVERRIDEWOW64}
procedure TVerifyApp.ChangeWow64Redirection(Active: boolean);
var OK: boolean;
begin
  if not Is64BitWindows then
    Exit;
  if Active then
    if fWow64RedirectionOldValueDONOTCHANGE <> nil then
      OK := Wow64RevertWow64FsRedirection(fWow64RedirectionOldValueDONOTCHANGE)
    else
      OK := False
  else
    OK := Wow64DisableWow64FsRedirection(fWow64RedirectionOldValueDONOTCHANGE);
  if not OK then
    Raise EOSError.CreateFmt('Failed to change WoW64 redirection state to %d.', [Integer(Active)]);
end;
{$ENDIF}

{$IFDEF CASESENSITIVESUPPORT}
{$IFDEF DELPHI6_UP}
  {$WARN UNIT_PLATFORM OFF}
  {$WARN SYMBOL_PLATFORM OFF}
{$ENDIF}

{$IFNDEF DELPHI6_UP}
  // Delphi 5 has an incorrect declaration of Windows.FindFirstFileEx
  function FindFirstFileEx(lpFileName: PChar; fInfoLevelId: TFindexInfoLevels;
    lpFindFileData: Pointer; fSearchOp: TFindexSearchOps; lpSearchFilter: Pointer;
    dwAdditionalFlags: DWORD): THandle; stdcall; external kernel32 name 'FindFirstFileExA';
{$ENDIF}

  // Copied from SysUtils
  function FindMatchingFile(var F: TSearchRec): Integer;
  var
    LocalFileTime: TFileTime;
  begin
    with F do
    begin
      while FindData.dwFileAttributes and ExcludeAttr <> 0 do
        if not FindNextFile(FindHandle, FindData) then
        begin
          Result := GetLastError;
          Exit;
        end;
      FileTimeToLocalFileTime(FindData.ftLastWriteTime, LocalFileTime);
      FileTimeToDosDateTime(LocalFileTime, LongRec(Time).Hi,
        LongRec(Time).Lo);
      Size := FindData.nFileSizeLow;
      Attr := FindData.dwFileAttributes;
      Name := FindData.cFileName;
    end;
    Result := 0;
  end;

function TVerifyApp.FindFirst(const Path: string; Attr: Integer; var SR: TSearchRec): Integer;
begin
  // This will only set-up the contents of SR
  Result := SysUtils.FindFirst(Path, Attr, SR);
  if CaseSensitive then
    begin
    Writeln('Case sensitive search for ' + Path);
    SysUtils.FindClose(SR);
    // Now I will create the search handle with required parameters
    SR.FindHandle := FindFirstFileEx(PChar(Path), FindExInfoStandard, @SR.FindData, FindExSearchNameMatch, nil, FIND_FIRST_EX_CASE_SENSITIVE);
    if SR.FindHandle <> INVALID_HANDLE_VALUE then
      begin
      Result := FindMatchingFile(SR);
      if Result <> 0 then SysUtils.FindClose(SR);
      end
    else
      Result := GetLastError;
    end;
end;

{$IFDEF DELPHI6_UP}
  {$WARN UNIT_PLATFORM ON}
  {$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
{$ENDIF}

function TVerifyApp.ProcessFiles(const BasePath, Path: string; ProcessFileFunction: TProcessFileFunction; Tag: integer): integer;
const ITEM_IS_FILE = 0;
      ITEM_IS_DIRECTORY = 1;
var SR: TSearchRec;
    List: TStringList;
    i, Code: integer;
begin
  List := TStringList.Create;
  try
    if FindFirst(BasePath + Path + '*.*', faAnyFile, SR) = 0 then
      try
        repeat
          CheckControlEvent;
          {$IFDEF WIN32}
          {$IFDEF DELPHI2009_UP}
          {$WARN SYMBOL_PLATFORM OFF}
          {$ENDIF}
          if (not Junctions) and Longbool(SR.FindData.dwFileAttributes and FILE_ATTRIBUTE_REPARSE_POINT) then
            begin
            // Skip
            end
          else
          {$IFDEF DELPHI2009_UP}
          {$WARN SYMBOL_PLATFORM ON}
          {$ENDIF}
          {$ENDIF}
          if Longbool(SR.Attr and faDirectory) then
            begin
            if Recursive and (SR.Name <> '.') and (SR.Name <> '..') then
              List.AddObject(SR.Name, TObject(ITEM_IS_DIRECTORY));
            end
          {$IFNDEF DELPHI2009_UP}
          else if Longbool(SR.Attr and faVolumeID) then
            begin
            // Nothing to do
            end
          {$ENDIF}
          else
            begin
            List.AddObject(SR.Name, TObject(ITEM_IS_FILE));
            end;
        until FindNext(SR) <> 0;
      finally
        SysUtils.FindClose(SR);
        end;
    List.Sort;
    Result := EXITCODE_OK;
    for i := 0 to Pred(List.Count) do
      begin
      CheckControlEvent;
      Code := EXITCODE_OK;
      if List.Objects[i] = TObject(ITEM_IS_DIRECTORY) then
        Code := ProcessFiles(BasePath, Path + List[i] + '\', ProcessFileFunction, Tag)
      else if IsFileWanted(Path + List[i]) then
        begin
        if ShowProgress then
          Write(Path + List[i]);
        Code := ProcessFileFunction(BasePath, Path + List[i], Tag);
        if ShowProgress then
          Writeln(PROGRESS_CLEAR);
        end;
      if Code <> EXITCODE_OK then
        Result := Code;
      end;
  finally
    List.Free;
    end;
end;

procedure TVerifyApp.HashingProgress(Sender: TObject; Done, Total: int64);
var Ticks: DWORD;
    PromileDone: int64;
begin
  CheckControlEvent;
  if ShowProgress then
    begin
    Ticks := GetTickCount;
    if (Done = 0) or ((Ticks - fLastProgressTicks) >= PROGRESS_EVENT_TICKS) then
      begin
      fLastProgressTicks := Ticks;
      PromileDone := 1000 * Done div Total;
      Write(Format(' %3.3s.%d%%'#8#8#8#8#8#8#8, [IntToStr(PromileDone div 10), PromileDone mod 10]));
      end;
    end;
end;

function CreateRegEx(const Pattern: string): TRegExp;
begin
  Result := RegExCreate(Pattern, [rcoSingleLine, rcoIgnoreCase]);
end;

function CreateRegExFromFileMask(Mask: string): TRegExp;
var i: integer;
    Pattern: string;
begin
  Mask := ExtractFileName(Mask); // Ignore path
  Pattern := '';
  for i := 1 to Length(Mask) do
    case Mask[i] of
      '*':
        Pattern := Pattern + '.*';
      '?':
        Pattern := Pattern + '.';
      ']':
        Pattern := Pattern + '\]';
      '^':
        Pattern := Pattern + '\^';
      '\':
        Pattern := Pattern + '\\';
      else
        Pattern := Pattern + '[' + Mask[i] + ']';
      end;
  Result := CreateRegEx('^' + Pattern + '$');
end;

procedure TVerifyApp.ClearIncludeExcludeMasks;
var i: integer;
begin
  for i := 0 to Pred(IncludeExcludeList.Count) do
    RegExFree(TRegExp(IncludeExcludeList.Objects[i]));
  IncludeExcludeList.Clear;
end;

procedure TVerifyApp.AddIncludeExcludeMask(const Mask: string; MaskType: TIncludeExcludeMask);
begin
  if IncludeExcludeList.Count = 0 then
    case MaskType of
      iemInclude, iemIncludeRegExp:
        IncludeExcludeList.AddObject(INCEXCTYPE_EXCLUDE, CreateRegEx('.*'));
      iemExclude, iemExcludeRegExp:
        IncludeExcludeList.AddObject(INCEXCTYPE_INCLUDE, CreateRegEx('.*'));
      end;
  case MaskType of
    iemInclude:
      IncludeExcludeList.AddObject(INCEXCTYPE_INCLUDE + Mask, CreateRegExFromFileMask(Mask));
    iemExclude:
      IncludeExcludeList.AddObject(INCEXCTYPE_EXCLUDE + Mask, CreateRegExFromFileMask(Mask));
    iemIncludeRegExp:
      IncludeExcludeList.AddObject(INCEXCTYPE_INCLUDE + Mask, CreateRegEx(Mask));
    iemExcludeRegExp:
      IncludeExcludeList.AddObject(INCEXCTYPE_EXCLUDE + Mask, CreateRegEx(Mask));
    end;
end;

procedure TVerifyApp.ApplyIncludeExcludeToHashList(HashList: THashList);
var i: integer;
begin
  for i := Pred(HashList.Count) downto 0 do
    if not IsFileWanted(HashList[i].Path) then
      HashList.Delete(i);
end;

function TVerifyApp.IsFileWanted(const Path: string): boolean;
var i: integer;
    s: string;
begin
  Result := True;
  for i := 0 to Pred(IncludeExcludeList.Count) do
    if TRegExp(IncludeExcludeList.Objects[i]).Match(Path) then
      begin
      s := IncludeExcludeList[i];
      if (s <> '') and (s[1] = INCEXCTYPE_INCLUDE) then
        Result := True
      else if (s <> '') and (s[1] = INCEXCTYPE_EXCLUDE) then
        Result := False;
      end;
end;

procedure TVerifyApp.DoConsoleCtrl(const CtrlType: DWORD; var Handled: boolean);
begin
  if ControlEvent = CTRL_NO_EVENT then // Regular processing on double Ctrl+Break
    if CtrlType in [CTRL_C_EVENT, CTRL_BREAK_EVENT, CTRL_CLOSE_EVENT] then
      begin
      ControlEvent := CtrlType;
      Handled := True;
      end;
  inherited;
end;

procedure TVerifyApp.CheckControlEvent;
begin
  case ControlEvent of
    CTRL_C_EVENT:
      begin
      ControlEvent := CTRL_NO_EVENT;
      Raise EControlC.Create('Ctrl+C pressed.');
      end;
    CTRL_BREAK_EVENT:
      begin
      ControlEvent := CTRL_NO_EVENT;
      Raise EControlC.Create('Ctrl+Break pressed.');
      end;
    CTRL_CLOSE_EVENT:
      begin
      ControlEvent := CTRL_NO_EVENT;
      Raise EControlC.Create('Application close requested.');
      end;
    end;
end;

function TVerifyApp.HandleCtrlBreakError(const Msg: string): boolean;
var s: string;
begin
  ShowError(Msg);
  WriteColored(ccWhite, 'Do you want to quit ('); WriteColored(ccLightRed, 'Q');
  WriteColored(ccWhite, ') now or show results first (anything else)?');
  Readln(s);
  Result := (s <> 'Q') and (s <> 'q');
end;

end.
